==> saw-0.1.1.tar.gz: saw-0.1.1/saw/saw.py <==
# -*- coding: utf-8 *-*

import os
import sys

from contextlib import closing
from sqlalchemy import types
from sqlalchemy import schema
from sqlalchemy import Column
from sqlalchemy import MetaData
from sqlalchemy import create_engine
from sqlalchemy.orm import mapper
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative import declarative_base


class Table(schema.Table):
    """A table."""

    _name = None
    metadata = None

    def __init__(self, name, *args, **kwargs):
        self.name = name
        super(Table, self).__init__(name, self.metadata, *args, **kwargs)

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name.lower()

    name = property(fget=get_name, fset=set_name)


class DefaultSession(object):
    """A default session container."""

    def __get__(self, instance, owner):
        return self._session

    def __set__(self, name, value):
        self._session = value


class Rows(object):
    """Util class to work with comon operations like:
            - insert
            - deleting
            - updating
            - selecting/quering
    """

    def __init__(self, db):
        """db is the DB instance"""
        self.db = db

    defaultsession = DefaultSession()

    def insert(self, obj, with_session=None, returning_id=None):
        """Insert a new register, optionally giving a session object."""

        session = with_session if with_session else self.defaultsession()

        try:
            session.add(obj)
            session.commit()
            if returning_id:
                returning_id = getattr(obj, returning_id)
        except Exception as exc:
            session.rollback()
            raise exc
        finally:
            session.close()

        return returning_id


    def delete(self, obj, with_session=None):
        """Removes a register, optionally giving a session object."""

        session = with_session if with_session else self.defaultsession()

        try:
            session.delete(obj)
            session.commit()
        except Exception as exc:
            session.rollback()
            raise exc
        finally:
            session.close()

    def update(self, obj, with_session=None):
        """Updates a register, optionally giving a session object."""

        session = with_session if with_session else self.defaultsession()

        try:
            session.merge(obj)
            session.commit()
        except Exception as exc:
            session.rollback()
            raise exc
        finally:
            session.close()

    def query(self, obj, with_session=None):
        """Create and returns a query object."""

        with self.db.autosession() as session:
            return session.query(obj)


class LocalTable(object):

    def __init__(self, db):
        self.attrdict = db.mappedtables

    def __getattribute__(self, attr):
        adict = object.__getattribute__(self, 'attrdict')
        if attr in adict:
            return adict[attr]
        else:
            return object.__getattribute__(self, attr)

    def __setattribute__(self, name, value):
        adict = object.__getattribute__(self, 'attrdict')
        if name in adict:
            adict[name] = value
        else:
            object.__setattr__(self, name, value)


class MetaTool(object):

    def __init__(self, class_, metas=None):
        ''' MetaTool - Invoque it when you are in danger!
            Parametros:
                class_: Instancia de la clase DB
                metas: tupla con instancias de MetaData
        '''

        if not isinstance(class_, DB):
            raise ValueError('class_ must be a saw.DB instance.')
        if metas:
            if not isinstance(metas, tuple):
                raise ValueError('metas must be a tuple of MetaDatas')

        self.metas = metas
        self.class_ = class_
        self.class_.table = LocalTable(self.class_)

    def has_primary_key(self, table):
        ''' Retorns a boolean if a table has or not has a primary key.
        '''
        if not len(table.primary_key):
            return False

        return True

    def create_fake_primary_key_on(self, mtable):
        ''' Create a fake primary key on metadata table.
        '''

        engine = self.class_.engine
        ddlc = engine.dialect.ddl_compiler(engine.dialect, None)

        fake_pk = Column('_id_pk', types.Integer, server_default="0",
                primary_key=True, autoincrement=True)
        mtable.append_column(fake_pk)

        try:
            specs = ddlc.get_column_specification(
                getattr(mtable.columns, '_id_pk'))
            sql = 'ALTER TABLE %s ADD %s' % (mtable.name, specs)
            engine.execute(sql)
        except:  # LOG ME NICELY
            pass

        return mtable

    def create_mapped_tables(self):

        tablemapping = self.class_.mappedtables

        if self.metas is None:
            return

        for meta in self.metas:
            for tname in meta.tables.iterkeys():
                tname = str(tname)  # puede ser unicode, por eso casteamos
                mtable = meta.tables[tname]  # stands for metadata table
                if not self.has_primary_key(mtable):
                    mtable = self.create_fake_primary_key_on(mtable)
                tablemapping[tname] = type(tname, (object,), dict())
                try:
                    mapper(tablemapping[tname], meta.tables[tname])
                except Exception, exc:
                    sys.stdout.write(exc.message)


class DB(object):

    def __init__(self, name, engine, autoload=False, **kwargs):
        """A glorius Database.

        Arguments:
            name(str): name from the database.
            engine(str): name from the database engine, like:
                'sqlite', 'mysql', 'postgresql', 'oracle'
            autoload(bool): Reflect the database structure. It means that
                            automagically sqlalchemy guess all your tables
                            and coumns structures without the need of declaring
                            models.

        Optional arguments (depends on options given above):
            username='username'
            password='secret'
            host='127.0.0.1'

        If we use 'sqlite' as engine, we can give those extra cute options:
            in_memory(True)  If we want to create an in memory sqlite database,
                             really usefull for tests tasks..
            destdir = If its not in_memory, set this to the path where you want
                      the sqlite database file be created or the path to the
                      sqlite file that you want to open.

        Deep tweaks, optional engine parameters:
            pool_size=20  # how many connections to keep alive/established
            pool_recycle=3600  # time in seconds to recycle the connections
        """

        if not isinstance(engine, Engine):
            if engine == "sqlite":
                if kwargs.get('in_memory'):
                    engine = create_engine('sqlite:///:memory:')
                else:
                    if not os.path.isfile(name):
                        destdir = kwargs.get(
                            'destdir', os.path.abspath(os.path.curdir))
                        name = os.path.join(destdir, name)
                    engine = create_engine(
                        'sqlite:///{database}'.format(database=name)
                    )
            else:
                password = kwargs.get('password')
                if not password:
                    uri = '{engine}://{username}@{host}/{name}'.format(
                        engine=engine, username=kwargs['username'],
                        host=kwargs['host'], name=name)
                else:
                    uri = '{engine}://{username}:{password}@{host}/{name}'
                    uri = uri.format(
                        engine=engine, username=kwargs['username'],
                        password=password, host=kwargs['host'], name=name)

                engine = create_engine(
                    uri,
                    pool_size=kwargs.get("pool_size", 20),
                    pool_recycle=kwargs.get("pool_recycle", 3600))

        self.__NAME__ = name        # name from the database
        self.__TABLES__ = dict()
        self.__ENGINE__ = engine    # create_engine object
        self.__META__ = MetaData(bind=engine)  # metadata about tables and their structure
        self.__SESSION__ = sessionmaker(engine)

        if autoload:
            self.__META__.reflect(bind=engine)

        self.metatool = MetaTool(class_=self, metas=(self.metadata,))
        self.metatool.create_mapped_tables()

        self.rows = Rows(db=self)
        self.rows.defaultsession = self.session
        self.register = self.rows  # TODO: self.register will be deprecated.

    @property
    def Model(self):
        return declarative_base(metadata=self.metadata)

    @property
    def name(self):
        return self.__NAME__

    @property
    def tables(self):
        return self.metadata.tables

    @property
    def mappedtables(self):
        return self.__TABLES__

    @property
    def engine(self):
        return self.__ENGINE__

    @property
    def metadata(self):
        return self.__META__

    @property
    def session(self):
        return self.__SESSION__

    def autosession(self):
        return closing(self.session())

    def create(self):
        """Create defined tables."""
        self.metadata.create_all(self.engine)
        self.metatool.create_mapped_tables()

    def drop(self):
        """Drop all tables, really dangerous!"""
        self.metadata.drop_all(self.engine)

==> saw-0.1.2.tar.gz: saw-0.1.2/saw/saw.py <==
# -*- coding: utf-8 *-*

import os
import sys
import threading

from contextlib import contextmanager
from sqlalchemy import MetaData
from sqlalchemy import create_engine
from sqlalchemy.orm import mapper
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative import declarative_base


class LocalTable(object):

    def __init__(self, db):
        self.db = db
        self.attrdict = db.mappedtables

    def __getattribute__(self, attr):
        adict = object.__getattribute__(self, 'attrdict')
        if attr in adict:
            return adict[attr]
        else:
            return object.__getattribute__(self, attr)

    def __setattribute__(self, name, value):
        adict = object.__getattribute__(self, 'attrdict')
        if name in adict:
            adict[name] = value
        else:
            object.__setattr__(self, name, value)

    def __iter__(self):
        return self.db.metadata.tables.iterkeys()

    def keys(self):
        return self.db.metadata.tables.keys()


class SessionManager(object):
    """Handles the sessions in a multi-threaded environment.
    """

    def __init__(self, engine):
        self.engine = engine
        self._sessions = {}

    def get_session(self):
        """Return a session for this thread."""
        thread_id = threading.current_thread().ident
        try:
            return self._sessions[thread_id]
        except KeyError:
            pass

        conn = self.engine.connect()
        Session = sessionmaker()
        session = Session(bind=conn)
        self._sessions[thread_id] = session
        return session


class DB(object):

    def __init__(self, name, engine, autoload=False, **kwargs):
        """A glorius Database.

        Arguments:
            name(str): name from the database.
            engine(str): name from the database engine, like:
                'sqlite', 'mysql', 'postgresql', 'oracle'
            autoload(bool): Reflect the database structure. It means that
                            automagically sqlalchemy guess all your tables
                            and coumns structures without the need of declaring
                            models.

        Optional arguments (depends on options given above):
            username='username'
            password='secret'
            host='127.0.0.1'

        If we use 'sqlite' as engine, we can give those extra cute options:
            in_memory(True)  If we want to create an in memory sqlite database,
                             really usefull for tests tasks..
            destdir = If its not in_memory, set this to the path where you want
                      the sqlite database file be created or the path to the
                      sqlite file that you want to open.

        Deep tweaks, optional engine parameters:
            pool_size=20  # how many connections to keep alive/established
            pool_recycle=3600  # time in seconds to recycle the connections
        """

        if not isinstance(engine, Engine):
            if engine == "sqlite":
                if kwargs.get('in_memory'):
                    engine = create_engine('sqlite:///:memory:')
                else:
                    if not os.path.isfile(name):
                        destdir = kwargs.get(
                            'destdir', os.path.abspath(os.path.curdir))
                        name = os.path.join(destdir, name)
                    engine = create_engine(
                        'sqlite:///{database}'.format(database=name)
                    )
            else:
                password = kwargs.get('password')
                if not password:
                    uri = '{engine}://{username}@{host}/{name}'.format(
                        engine=engine, username=kwargs['username'],
                        host=kwargs['host'], name=name)
                else:
                    uri = '{engine}://{username}:{password}@{host}/{name}'
                    uri = uri.format(
                        engine=engine, username=kwargs['username'],
                        password=password, host=kwargs['host'], name=name)

                engine = create_engine(
                    uri,
                    pool_size=kwargs.get("pool_size", 20),
                    pool_recycle=kwargs.get("pool_recycle", 3600))

        self._name = name
        self._engine = engine
        self._tables = dict()
        self._metadata = MetaData(bind=engine)
        self._session_manager = SessionManager(engine)

        if autoload:
            self._metadata.reflect(bind=engine)

        self.tables = LocalTable(self)
        self.create_mapped_tables()

    @property
    def Model(self):
        return declarative_base(metadata=self.metadata)

    @property
    def name(self):
        return self._name_

    @property
    def mappedtables(self):
        return self._tables

    @property
    def engine(self):
        return self._engine

    @property
    def metadata(self):
        return self._metadata

    @property
    def get_session(self):
        session_manager = self._session_manager
        return session_manager.get_session

    @property
    @contextmanager
    def query(self):
        """session.query()"""
        session = self.get_session()
        yield session.query
        session.close()

    @property
    @contextmanager
    def session(self):
        """Autoclosing session"""
        session = self.get_session()
        yield session
        session.close()

    def create(self):
        """Create defined tables."""
        self.metadata.create_all(self.engine)
        self.create_mapped_tables()

    def insert(self, obj, returning_id=None):
        """Insert a new object."""

        session = self.get_session()

        try:
            session.add(obj)
            session.commit()
            if returning_id:
                returning_id = getattr(obj, returning_id)
        except Exception as exc:
            session.rollback()
            raise exc
        finally:
            session.close()

        return returning_id

    def delete(self, obj):
        """Removes an object."""

        session = self.get_session()

        try:
            session.delete(obj)
            session.commit()
        except Exception as exc:
            session.rollback()
            raise exc
        finally:
            session.close()

    def update(self, obj):
        """Updates an object"""

        session = self.get_session()

        try:
            session.merge(obj)
            session.commit()
        except Exception as exc:
            session.rollback()
            raise exc
        finally:
            session.close()

    def create_mapped_tables(self):

        tablemapping = self._tables
        meta = self.metadata

        for tname in meta.tables.iterkeys():
            tname = str(tname)
            tablemapping[tname] = type(tname, (object,), dict())
            try:
                mapper(tablemapping[tname], meta.tables[tname])
            except Exception, exc:
                sys.stdout.write(exc.message)
